/*
  The contents of this file are subject to the University of Utah Public
  License (the "License"); you may not use this file except in
  compliance with the License.
  
  Software distributed under the License is distributed on an "AS IS"
  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See
  the License for the specific language governing rights and limitations
  under the License.

  The Original Source Code is "teem", released March 23, 2001.
  
  The Original Source Code was developed by the University of Utah.
  Portions created by UNIVERSITY are Copyright (C) 2001, 1998 University
  of Utah. All Rights Reserved.
*/


#include "air.h"

/*
******** airFree()
**
** just to facilitate setting a newly free()'d pointer to NULL
**
** I may regret doing the check on non-NULL-ity before sending the argument
** to free()- the caller could get reckless about sending already NULL
** pointers.
*/
void *
airFree(void *ptr) {

  if (ptr)
    free(ptr);
  return NULL;
}

/*
******** airStrdup()
**
** because they didn't put strdup() in ANSI ...
*/
char *
airStrdup(char *s) {
  char *ret;

  if (!s) {
    ret = NULL;
  }
  else {
    ret = malloc(strlen(s)+1);
    if (ret) {
      strcpy(ret, s);
    }
  }
  return ret;
}


/*
******* airDoneStr()
**
** dinky little utility for generating progress messages of the form
** "  1.9%" or " 35.3%" or  "100.0%"
** The message will ALWAYS be six characters, and will ALWAYS be
** preceeded by six backspaces.  You pass in a string to print into
** and it had better be allocated for at least 6+6+1 = 13 chars
*/
char *
airDoneStr(float start, float here, float end, char *str) {
  int perc;
  static int len = -1;

  if (str) {
    if (end != start)
      perc = (int)(1000*(here-start)/(end-start) + 0.5);
    else
      perc = 1000;
    if (perc < 0) {
      sprintf(str, "\b\b\b\b\b\b ---- ");
    } else if (perc < 1000) {
      sprintf(str, "\b\b\b\b\b\b% 3d.%d%%", perc/10, perc%10);
    }
    else if (perc == 1000) {
      /* the "% 3d" formatting sequence should have taken care
	 of this, but whatever */
      sprintf(str, "\b\b\b\b\b\b100.0%%");
    }
    else {
      sprintf(str, "\b\b\b\b\b\b  done");
    }
  }

  if (-1 == len) {
    len = strlen(str);
  }
  else {
    if (len != strlen(str)) {
      printf("len(\\b\\b\\b\\b\\b\\b% 3d.%d%%) != %d\n", 
	      perc/10, perc%10, len);
      exit(1);
    }
  }
  return(str);
}

/*
******** airTime()
**
** returns current time in milliseconds as a double
*/
double
airTime() {
  struct timeval tv;
  struct timezone tz;

  gettimeofday(&tv, &tz); 
  return((double)(tv.tv_sec + tv.tv_usec/1000000.0));
}

/*
******** airLog2()
**
** silly little function which returns 1+log_2(n) if n is a power of 2,
** or 0 otherwise
*/
int 
airLog2(float n) {
  int ret;

  if (!AIR_EXISTS(n))
    return 0;
  if (n == 1.0)
    return 1;
  if (n < 2)
    return 0;
  /* else */
  ret = airLog2(n/2.0);
  if (ret) {
    ++ret;
  }
  return ret;
}
