/*
  The contents of this file are subject to the University of Utah Public
  License (the "License"); you may not use this file except in
  compliance with the License.
  
  Software distributed under the License is distributed on an "AS IS"
  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied.  See
  the License for the specific language governing rights and limitations
  under the License.

  The Original Source Code is "teem", released March 23, 2001.
  
  The Original Source Code was developed by the University of Utah.
  Portions created by UNIVERSITY are Copyright (C) 2001, 1998 University
  of Utah. All Rights Reserved.
*/


#include "air.h"

#define _PARSE_STR_ARGS(type, format) type *out, char *_s, char *ct, int n
#define _PARSE_STR_BODY(type, format) \
  int i; \
  char *tmp, *s; \
  \
  /* if we got NULL, there's nothing to do */ \
  if (!(out && _s && ct)) \
    return 0; \
  \
  /* copy the input so that we don't change it */ \
  s = airStrdup(_s); \
  \
  /* keep calling strtok() until we have everything */ \
  for (i=0; i<=n-1; i++) { \
    tmp = strtok(i ? NULL : s, ct); \
    if (!tmp) { \
      free(s); \
      return i; \
    } \
    if (1 != sscanf(tmp, format, out+i)) { \
      free(s); \
      return i; \
    } \
  } \
  free(s); \
  return n; \

/*
******* airParseI(), airParseF(), airParseF(), airParseC()
**
** parse ints, floats, doubles, or single characters, from some
** given string "_s"; try to parse "n" of them, and put the 
** results in "out".
**
** Returns the number of things succesfully parsed- should be n; 
** there's been an error if return is < n.
** Does not use biff.
**
** int airParseStrI(int *out, char *_s, char *ct, int n)
** int airParseStrF(float *out, char *_s, char *ct, int n)
** int airParseStrD(double *out, char *_s, char *ct, int n)
** int airParseStrC(char *out, char *_s, char *ct, int n)
**
** WARNING: this calls strtok() (not thread-safe)
*/
int

airParseStrI(_PARSE_STR_ARGS(int, "%d"))     { _PARSE_STR_BODY(int, "%d") }

/*
airParseStrI(_PARSE_STR_ARGS(int, "%d"))     {
  int i;
  char *tmp, *s;
 
  / * if we got NULL, there's nothing to do * /
  if (!(out && _s && ct))
    return 0;
 
  / * copy the input so that we don't change it * /
  s = airStrdup(_s);
 
  / * keep calling strtok() until we have everything * /
  for (i=0; i<=n-1; i++) {
    tmp = strtok(i ? NULL : s, ct);
    if (!tmp) {
      free(s);
      return i;
    }
    if (1 != sscanf(tmp, "%d", out+i)) {
      free(s);
      return i;
    }
  }
  free(s);
  return n;
}
*/

int
airParseStrF(_PARSE_STR_ARGS(float, "%f"))   { _PARSE_STR_BODY(float, "%f") }

int
airParseStrD(_PARSE_STR_ARGS(double, "%lf")) { _PARSE_STR_BODY(double, "%lf") }

int
airParseStrC(_PARSE_STR_ARGS(char, "%c"))    { _PARSE_STR_BODY(char, "%c") }

int
airParseStrS(char **out, char *_s, char *ct, int n) {
  int i, j;
  char *tmp, *s;

  /* if we got NULL, there's nothing to do */
  if (!(out && _s && ct))
    return 0;

  /* copy the input so that we don't change it */
  s = airStrdup(_s);

  /* keep calling strtok() until we have everything */
  for (i=0; i<=n-1; i++) {
    tmp = strtok(i ? NULL : s, ct);
    if (!tmp) {
      free(s);
      return i;
    }
    out[i] = airStrdup(tmp);
    if (!out[i]) {
      for (j=0; j<=i-1; j++) 
	free(out[j]);
      free(s);
      return i;
    }
  }
  free(s);
  return n;
}
